from dataclasses import dataclass, field
from datetime import timedelta
from pathlib import Path
from typing import TypeVar
from zenlog import log

from numpy import nan

from mccode_antlr.reader import Registry

TestInstrExampleType = TypeVar('TestInstrExampleType', bound='TestInstrExample')


@dataclass
class TestInstrExample:
    """Holds parameters for a single instr file %Example: test case"""
    sourcefile: Path
    test_number: int
    parameter_values: str = ""
    detector: str = ""
    target_value: float = nan
    test_value: float = nan
    compiled: bool = False
    compile_time: timedelta = field(default_factory=timedelta)
    ran: bool = False
    run_time: timedelta = field(default_factory=timedelta)
    error_message: str = ""
    stdout: str = ""
    test_complete: bool = False

    def __post_init__(self):
        import re
        if '.instr' in self.parameter_values:
            rinstr = re.compile(r'[a-zA-Z][a-zA-Z0-9_\-]*\.instr')
            new_param = rinstr.sub('', self.parameter_values)
            log.info(f'Removing .instr from {self.parameter_values} now {new_param}')
            self.parameter_values = new_param
        if self.sourcefile.stem in self.parameter_values:
            new_param = self.parameter_values
            self.parameter_values = new_param.replace(self.sourcefile.stem, '')
            log.info(f'Removing {self.sourcefile.stem} from {new_param} now {self.parameter_values}')
        if '-n' in self.parameter_values:
            rn = re.compile(r'-n[0-9eE\.]*')
            new_param = rn.sub('', self.parameter_values)
            log.info(f'Removing -n from {self.parameter_values}, now {new_param}')
            self.parameter_values = new_param

    @classmethod
    def list_from_file(cls, filename: Path) -> list[TestInstrExampleType]:
        import re
        rspec = re.compile(f'%Example:([^\n]*)Detector:([^\n]*)_I=([0-9.+-e]+)')
        with open(filename, 'r') as file:
            contents = file.read()
            matches = rspec.findall(contents)
        if len(matches):
            groups = [(m[0].strip(), m[1].strip(), float(m[2].strip())) for m in matches]
            return [cls(filename, i+1, p, d, t) for i, (p, d, t) in enumerate(groups)]
        return [cls(filename, -1)]

    def get_display_name(self):
        return f'{self.sourcefile.stem}_{self.test_number}' if self.test_number > 1 else self.sourcefile.stem

    def get_json_repr(self):
        jr = dict(displayname=self.get_display_name(),
                  sourcefile=self.sourcefile.as_posix(),
                  testnb=self.test_number,
                  parvals=self.parameter_values,
                  detector=self.detector,
                  targetval=self.target_value,
                  testval=self.test_value,
                  compiled=self.compiled,
                  compiletime=str(self.compile_time),
                  didrun=self.ran,
                  runtime=str(self.run_time),
                  errmsg=self.error_message,
                  stdout=self.stdout,
                  complete=self.test_complete
                  )
        return jr

    def save(self, path: Path):
        import json
        with open(path.joinpath(self.get_display_name() + '.json'), 'w') as file:
            json.dump(self.get_json_repr(), file)


@dataclass
class LineLogger:
    lines: list = field(default_factory=list)

    def logline(self, line: str):
        self.lines.append(line)

    def save(self, filename: Path):
        with open(filename, 'w') as file:
            file.write('\n'.join(self.lines))
            file.write('\n')

    def find(self, search: str):
        import re
        for line in self.lines:
            if re.search(search, line):
                return True
        return False


def get_cpu_name():
    from platform import system
    from subprocess import check_output
    if system() == "Windows":
        from platform import processor
        return processor()
    elif system() == "Darwin":
        from os import environ, pathsep
        environ['PATH'] = environ['PATH'] + pathsep + '/usr/sbin'
        command = "sysctl -n machdep.cpu.brand_string"
        return check_output(command).strip()
    elif system() == "Linux":
        command = "cat /proc/cpuinfo"
        all_info = check_output(command, shell=True).decode().strip()
        for line in all_info.split("\n"):
            if "model name" in line:
                from re import sub
                return sub(".*model name.*:", "", line, 1).strip()
    return ""


def get_nvidia_gpu_name():
    """Identify GPU (Uses GPUtil which only knows about Nvidia GPUs)"""
    import GPUtil
    gpus = GPUtil.getGPUs()
    names = ','.join(gpu.name for gpu in gpus)
    return names if len(names) else "none"


def get_hostname():
    import socket
    return socket.gethostname()


def get_username():
    import os
    import pwd
    info = pwd.getpwuid(os.getuid())
    # Return the full name 'pw_gecos' if present, otherwise the username 'pw_name'
    return info[4] if len(info[4]) else info[0]


def _monitor_name_file_name_match(folder, monitor_name):
    import re
    look_for_filename = False
    with open(folder.joinpath('mccode_antlr.sim'), 'r') as file:
        lines = file.readlines()
    for line in lines:
        if re.match(rf"\s*component:\s*{monitor_name}$", line):
            look_for_filename = True
        if look_for_filename:
            m = re.match(r'\s*filename:\s+(.+)', line)
            if m:
                filename = folder.joinpath(m.group(1))
                if filename.is_file():
                    return filename
            if re.match('end data', line):
                # try the monitor named dat file
                zero_d_filename = folder.joinpath(monitor_name + '.dat')
                if zero_d_filename.is_file():
                    return zero_d_filename
                return None
    return None


def mccode_test_compiler(work_dir, file_path, target, registry, generator, dump, **kwargs):
    from pathlib import Path
    from mccode_antlr.reader import Reader
    from mccode_antlr.compiler.c import compile_instrument
    # only the provided (remote) registry should be necessary for test instruments
    reader = Reader(registries=[registry])
    inst = reader.get_instrument(file_path)
    output = Path(work_dir)
    config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                  embed_instrument_file=False, verbose=False)
    try:
        binary_path = compile_instrument(inst, target, output, generator=generator, config=config, dump_source=dump, **kwargs)
    except RuntimeError as err:
        log.critical(f'Failed to produce a binary for {file_path}')
        log.info(err)
        return False, Path()
    return True, binary_path


def mcstas_test_compiler(target, work_dir, file_path, dump, **kwargs):
    from mccode_antlr.reader import MCSTAS_REGISTRY
    from mccode_antlr.translators.target import MCSTAS_GENERATOR
    return mccode_test_compiler(work_dir, file_path, target, MCSTAS_REGISTRY, MCSTAS_GENERATOR, dump, **kwargs)


def mcxtrace_test_compiler(target, work_dir, file_path, dump, **kwargs):
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    from mccode_antlr.translators.target import MCXTRACE_GENERATOR
    return mccode_test_compiler(work_dir, file_path, target, MCXTRACE_REGISTRY, MCXTRACE_GENERATOR, dump, **kwargs)


def mccode_test_runner(target, binary_path, test_parameters: str, n_particles: int):
    import shutil
    from tempfile import mkdtemp
    from mccode_antlr.compiler.c import run_compiled_instrument

    output_path = Path(mkdtemp(prefix=binary_path.stem, dir=binary_path.parent.resolve()))
    # The McCode C runtime won't use an existing directory ...
    shutil.rmtree(output_path)
    # common default parameters: output directory, seed value, number of particles
    parameters = f'--dir {output_path} -s 1000 -n {n_particles} ' + test_parameters
    try:
        output = run_compiled_instrument(binary_path, target, parameters, capture=True)
    except RuntimeError as error:
        return False, str(error), output_path
    return True, output, output_path


def mcstas_test_runner(target, binary_path, test_parameters: str, n_particles: int):
    # If NeXus output is requested and the InstrumentDescriptionFile is needed, run a different script entirely...
    #   TODO think about actually doing this?
    # if target.flags & CBinaryTarget.Type.nexus and idf_required:
    #     run([config['idfgen'].get(str), str(binary), *options])
    return mccode_test_runner(target, binary_path, test_parameters, n_particles)


def mcxtrace_test_runner(target, binary_path, test_parameters: str, n_particles: int):
    return mccode_test_runner(target, binary_path, test_parameters, n_particles)


def mccode_test(compiler, runner, mpi: int = None, acc: bool = False, nexus: bool = False, **kwargs):
    """Specialize the compiler and runner based on requested number of processors and use of OpenACC, then test"""
    from functools import partial
    from mccode_antlr.compiler.c import CBinaryTarget
    target = CBinaryTarget(mpi=mpi is not None, acc=acc, count=1 if mpi is None else mpi, nexus=nexus)
    return _mccode_test(partial(compiler, target), partial(runner, target), **kwargs)


def mcstas_test(search_pattern=None, instr_count=None, skip_non_test: bool = False,
                mpi: int = None, acc: bool = False, nexus: bool = False, n_particles: int = 1000,
                dump: bool = False, workdir: Path = None):
    from mccode_antlr.reader import MCSTAS_REGISTRY as REGISTRY
    return mccode_test(mcstas_test_compiler, mcstas_test_runner, mpi, acc, nexus, registry=REGISTRY,
                       search_pattern=search_pattern, instr_count=instr_count, skip_non_test=skip_non_test,
                       n_particles=n_particles, dump=dump, workdir=workdir)


def mcxtrace_test(search_pattern=None, instr_count=None, skip_non_test: bool = False,
                  mpi: int = None, acc: bool = False, nexus: bool = False, n_particles: int = 1000,
                  dump: bool = False, workdir: Path = None):
    from mccode_antlr.reader import MCXTRACE_REGISTRY as REGISTRY
    return mccode_test(mcxtrace_test_compiler, mcxtrace_test_runner, mpi, acc, nexus, registry=REGISTRY,
                       search_pattern=search_pattern, instr_count=instr_count, skip_non_test=skip_non_test,
                       n_particles=n_particles, dump=dump, workdir=workdir)


def _mccode_test(compiler, runner, registry: Registry, search_pattern=None, instr_count=None,
                skip_non_test: bool = False, n_particles: int = 1000, workdir: Path = None, dump: bool=False):
    """Run McCode instrument test(s) using the provided `compiler` and `runner` functions to handle those tasks.

    Each takes the expected working directory name (a string) and the path to the instr file as input.
    The compiler should return only a boolean success indication.
    The runner should return a boolean success indicator, the contents of the run standard output, and the directory
    (if any) used to store run output files -- these files are intended to reside below the working directory and will
    be automatically removed by this function.

    Parameters:
        compiler: a function taking a string and Path and producing the necessary runtime executable
        runner: a function taking a string and Path and running the executable, storing its output
        registry: a Registry object which knows where to find all 'examples/*' instr files
        search_pattern: an optional positive filter for the registered 'examples/*' instr files
        instr_count: an optional positive count of filtered instr files to include
        skip_non_test: a flag to control whether instr files which do not specify test cases should be compiled and run
    """
    import re
    # import logging
    from zenlog import log as logging
    import tempfile
    from datetime import datetime
    logging.info(f"Finding test instruments in: {registry}")
    # we _require_ that all test instr files are in a folder called "examples" ...
    filenames = registry.match(re.compile('examples'))
    # allow the user to limit which test cases can be found via a regular expression search
    if search_pattern is not None:
        if not isinstance(search_pattern, re.Pattern):
            search_pattern = re.compile(search_pattern)
        filenames = [x for x in filenames if search_pattern.search(x) is not None]
    # the examples folder has at least one script in it:
    filenames = [x for x in filenames if Path(x).suffix == '.instr']
    # optionally test case count limiter:
    if instr_count is not None and isinstance(instr_count, int) and instr_count > 0:
        filenames = filenames[:instr_count]
    # grab full paths to each file (in case this is a remote registry and they need to be fetched)
    filepaths = [registry.path(filename) for filename in filenames]

    longest_name = max(len(x.stem) for x in filepaths)

    # Build test objects from the '%Example:' line(s) in each file:
    tests = [x for filepath in filepaths for x in TestInstrExample.list_from_file(filepath)]

    binaries = dict()

    # By default, use a temporary directory -- but allow the user to specify someplace else:
    created = False
    if workdir is None:
        created = True
        workdir = tempfile.mkdtemp()
    if not isinstance(workdir, Path):
        workdir = Path(workdir)
    if not workdir.exists():
        created = True
        workdir.mkdir(parents=True)
    if not workdir.is_dir():
        raise RuntimeError(f'Expected {workdir} to be a directory, but it is not.')

    # Move into a temporary directory to ensure compilation/run times are from scratch

    logging.info("Compiling instruments")
    for test in tests:
        print(test)
        if test.sourcefile not in binaries and (test.test_number != 0 or not skip_non_test):
            t1 = datetime.now()
            binaries[test.sourcefile] = compiler(workdir, test.sourcefile, dump=dump, recompile=True, replace=False)
            test.compile_time = datetime.now() - t1
            test.compiled = True

            if binaries[test.sourcefile][0]:
                logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: {test.compile_time}')
            else:
                logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: COMPILE ERROR')
        else:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: COMPILE skipped')

    logging.info("Running tests")
    for test in tests:
        if not binaries.get(test.sourcefile, (False, None))[0]:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: No compile')
            continue
        if test.test_number < 1:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: No test')
            continue

        t1 = datetime.now()
        log.info(f'run {binaries[test.sourcefile][1]} -n {n_particles} {test.parameter_values}')
        test.ran, stdout, output_dir = runner(binaries[test.sourcefile][1], test.parameter_values, n_particles)
        test.run_time = datetime.now() - t1
        test.stdout = stdout.decode() if isinstance(stdout, bytes) else stdout

        if test.ran:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: {test.run_time}')
        else:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: RUNTIME ERROR')
            logging.debug(test.stdout)
            test.error_message = test.stdout
            continue

        detector_output = _monitor_name_file_name_match(output_dir, test.detector)
        logging.debug(f'Detector output stored in {detector_output}')
        test.test_value = - 1
        if detector_output is None or not detector_output.is_file():
            rdet = re.compile(f'{test.detector}_I= ([0-9+-eE.]+)')
            matches = rdet.findall(test.stdout)
            if len(matches):
                test.test_value = float(matches[0][0])
        else:
            with open(detector_output, 'r') as detector_file:
                detector_lines = detector_file.readlines()
            detector_pattern = re.compile('# values: ([0-9+-e.]+) ([0-9+-e.]+) ([0-9]+)')
            detector_matches = [detector_pattern.match(line) for line in detector_lines if detector_pattern.match(line)]
            if len(detector_matches):
                test.test_value = float(detector_matches[0].group(1))

        test.test_complete = True

    if created:
        import shutil
        shutil.rmtree(workdir)
    else:
        logging.info(f'Test compiled binaries and test case output directories are located under {workdir}.\n'
                     'You may wish to clean-up this directory to recover disk space.')

    # Since the output directories may no longer exist, save the test results ... here?
    for test in tests:
        test.save(Path())

    results = {t.get_display_name(): t.get_json_repr() for t in tests}
    # the 'ncount' and 'mpi' keys should be defined elsewhere
    results['_meta'] = {'date': datetime.now().isoformat(), 'hostname': get_hostname(),
                        'user': get_username(), 'cpu_type': get_cpu_name(), 'gpu_type': get_nvidia_gpu_name()}
    return results



def ispath(arg: str):
    from pathlib import Path
    return Path(arg)


def main(name, program):
    from argparse import ArgumentParser
    from zenlog import log
    parser = ArgumentParser(name, description=f'Test instrument compilation and runtime for {name}')
    parser.add_argument('-s', '--search', help='Regular expression positive filter for instrument names', default=None)
    parser.add_argument('-c', '--count', type=int, help='Maximum number of instruments to test', default=None)
    parser.add_argument('-k', '--skip', action='store_true', help='Skip instruments without test cases')
    parser.add_argument('--mpi', type=int, help='Number of mpi processes to use', default=None)
    parser.add_argument('--nexus', action='store_true', help='Whether to use NeXus output')
    parser.add_argument('-n', '--ncount', type=int, help='Number of particles to simulate per test', default=1000)
    parser.add_argument('-d','--dump', action='store_true', help='Output C source to file', default=False)
    parser.add_argument('-w','--workdir', type=ispath, help='Work directory, temporary if None', default=None)

    # logging.basicConfig(level=logging.DEBUG, format='{asctime} {levelname} {message}', style='{')
    # logging.basicConfig(level=logging.DEBUG)

    log.level('debug')

    args = parser.parse_args()
    results = program(search_pattern=args.search, instr_count=args.count, skip_non_test=args.skip, mpi=args.mpi,
                      nexus=args.nexus, n_particles=args.ncount, dump=args.dump, workdir=args.workdir)
    results['_meta']['mpi'] = args.mpi
    results['_meta']['ncount'] = args.ncount

def mcstas_test_main():
    return main('mcstas_test', mcstas_test)

def mcxtract_test_main():
    return main('mcxtrace_test', mcxtrace_test)


if __name__ == '__main__':
    mcstas_test_main()


from .builder import _ensure_antlr_files_up_to_date_on_import

_ensure_antlr_files_up_to_date_on_import('McComp', deps=('McCommon', 'cpp'), verbose=True)
_ensure_antlr_files_up_to_date_on_import('McInstr', deps=('McCommon', 'cpp'), verbose=True)
_ensure_antlr_files_up_to_date_on_import('C', verbose=True)



from zenlog import log
from ..grammar import CParser, CListener, McInstrParser
from ..instr import InstrVisitor
from ..common import Expr
from antlr4.error.ErrorListener import ErrorListener


def literal_string(ctx):
    start_token, stop_token = ctx.start, ctx.stop
    stream = start_token.getInputStream()
    return stream.getText(start_token.start, stop_token.stop)


class CErrorListener(ErrorListener):
    def __init__(self, source: str, pre=5, post=2):
        self.source = source
        self.pre = pre
        self.post = post

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        log.error(f'Syntax error in parsing {line},{column}')
        lines = self.source.split('\n')
        pre_lines = lines[line-self.pre:line]
        post_lines = lines[line:line+self.post]
        for line in pre_lines:
            log.info(line)
        log.error('~'*column + '^ ' + msg)
        for line in post_lines:
            log.info(line)


class DeclaresCListener(CListener):
    def __init__(self, typedefs: list = None, verbose=False):
        self.verbose = verbose
        self.typedefs = [] if typedefs is None else typedefs
        self.variables = dict()  # self.variables['name'] = type, value set by listeners
        self.last_type = None
        self.last_name = None
        self.last_value = None
        self.is_declaration = False
        self.is_typedef = False
        self.not_type_name = None
        self.was_type = False
        self.stored = False
        
    def debug(self, message):
        if self.verbose:
            log.debug(message)

    def enterDeclaration(self, ctx: CParser.DeclarationContext):
        self.debug('Enter declaration')
        self.is_declaration = True
        self.stored = False

    # Exit a parse tree produced by CParser#declaration.
    def exitDeclaration(self, ctx: CParser.DeclarationContext):
        if self.is_typedef:
            typename = self.not_type_name if self.last_name is None else self.last_name
            if typename is None:
                print(literal_string(ctx))
                print(self.last_type)
                print(self.last_name)
                print(self.last_value)
                print(self.not_type_name)
                return
            self.typedefs.append(typename.replace('*', '').strip())
            self.is_typedef = False
        elif not self.stored and self.not_type_name is not None and self.last_type.endswith(self.not_type_name):
            # this can only happen for declaration lines? where no initialization is performed.
            new_type = self.last_type[:-len(self.not_type_name)].strip()
            self.debug(f'Storing declaration for {self.not_type_name} = ({new_type}, {self.last_value})')
            if len(new_type):
                self.variables[self.not_type_name] = (new_type, self.last_value)
                self.stored = True
        self.not_type_name = None
        self.last_type = None
        self.last_name = None
        self.last_value = None
        self.is_declaration = False
        self.debug('Exit declaration')

    def enterDeclarationSpecifiers(self, ctx: CParser.DeclarationSpecifiersContext):
        self.last_type = literal_string(ctx)
        self.debug(f'Enter Declaration Specifier {self.last_type=}')

    def enterStorageClassSpecifier(self, ctx: CParser.StorageClassSpecifierContext):
        self.is_typedef = 'typedef' in literal_string(ctx)
        self.debug(f'Enter storage class specifier {self.is_typedef=}')

    def enterTypeSpecifier(self, ctx: CParser.TypeSpecifierContext):
        self.was_type = True
        self.debug(f'Enter type specifier')

    def enterTypedefName(self, ctx: CParser.TypedefNameContext):
        # Check here if this is a known typedef'd name or not :/
        self.was_type = literal_string(ctx) in self.typedefs
        if not self.was_type:
            # We erroneously ended up here thinking this is part of a type name.
            # It's not, but we need to save the string value _Somewhere_
            self.not_type_name = literal_string(ctx)
        self.debug(f'Enter typedef name {self.was_type=} {self.not_type_name=}')

    def exitInitDeclarator(self, ctx: CParser.InitDeclaratorContext):
        if not self.is_typedef:
            self.debug(f'Storing declaration for {self.last_name} = ({self.last_type}, {self.last_value})')
            self.variables[self.last_name] = (self.last_type, self.last_value)
            self.stored = True
            self.last_name = None
        self.last_value = None

    def enterDeclarator(self, ctx: CParser.DeclaratorContext):
        self.last_name = literal_string(ctx)
        self.debug(f'Enter declarator {self.last_name=}')

    def exitInitializer(self, ctx: CParser.InitializerContext):
        self.last_value = literal_string(ctx)
        self.debug(f'Exit initializer {self.last_value=}')


class EvalCVisitor(InstrVisitor):
    def __init__(self, found: dict = None, verbose=False):
        super().__init__(None, None)
        self.verbose = verbose
        self.assigned = {} if found is None else found
        self.prog = []

    def debug(self, message):
        if self.verbose:
            log.debug(message)

    def visitAssignment(self, ctx: McInstrParser.AssignmentContext):
        name = str(ctx.Identifier())
        value = self.visit(ctx.expr())
        self.prog.append((name, value))
        self.assigned[name] = value
        self.debug(f'{literal_string(ctx)} -> {name} = {value}')

    def visitExpressionIdentifier(self, ctx: McInstrParser.ExpressionIdentifierContext):
        from ..common import Expr
        # check if we already have a value for this identifier and insert it if we do:
        name = str(ctx.Identifier())
        if name in self.assigned:
            return self.assigned[name]
        return Expr.id(name)



def extract_c_declared_variables_and_defined_types(block: str, user_types: list = None, verbose=False):
    from antlr4 import InputStream, CommonTokenStream
    from antlr4 import ParseTreeWalker
    from ..grammar import CLexer
    # self.debug('Load block into ANTLR4 stream')
    # self.debug(f'{block}')
    stream = InputStream(block)
    # self.debug('Run lexer')
    lexer = CLexer(stream)
    # self.debug('Tokenize stream')
    tokens = CommonTokenStream(lexer)
    # self.debug('Parse tokens')
    parser = CParser(tokens)
    parser.addErrorListener(CErrorListener(block))
    # self.debug('Extract compilation unit tree')
    tree = parser.compilationUnit()
    # self.debug('Initialize listener')
    listener = DeclaresCListener(user_types, verbose=verbose)
    # self.debug('Initialize walker')
    walker = ParseTreeWalker()
    # self.debug('Walk the tree')
    walker.walk(listener, tree)
    return listener.variables, listener.typedefs


def extract_c_declared_variables(block: str, user_types: list = None, verbose=False):
    variables, types = extract_c_declared_variables_and_defined_types(block, user_types, verbose=verbose)
    return variables


def extract_c_defined_then_declared_variables(defined_in_block: str, declared_in_block):
    _, defined_in_types = extract_c_declared_variables_and_defined_types(defined_in_block)
    return extract_c_declared_variables(declared_in_block, user_types=defined_in_types)


def evaluate_c_defined_variables(variables: dict[str, str], initialized_in: str, known: dict[str, Expr] = None,
                                 verbose=False):
    """Evaluate individual statements from C-like source in an attempt to find values for the provided variables"""
    from antlr4 import InputStream, CommonTokenStream
    from ..grammar import McInstrLexer, McInstrParser
    from ..common import Expr, DataType
    lines = [f'{line};' for line in initialized_in.split(';') if len(line.strip())]
    found = {} if known is None else known
    for line in lines:
        parser = McInstrParser(CommonTokenStream(McInstrLexer(InputStream(line))))
        parser.addErrorListener((CErrorListener(line)))
        visitor = EvalCVisitor(found=found, verbose=verbose)
        visitor.visitAssignment(parser.assignment())
        found = visitor.assigned

    def get_found(name, data_type):
        expr = found.get(name, Expr.id(name)).simplify()
        if expr.is_singular:
            expr.data_type = DataType.from_name(data_type)
        return expr

    return {v: get_found(v, data_type) for v, data_type in variables.items()}


def _get_expr(type_name: str, initial_value: str) -> Expr:
    from ..common import DataType, Value
    expr = Expr(Value(None)) if initial_value is None else Expr.parse(initial_value)
    expr.data_type = DataType.from_name(type_name)
    return expr


def extract_c_declared_expressions(block: str, user_types: list = None, verbose=False) -> dict[str, Expr]:
    variables = extract_c_declared_variables(block, user_types, verbose=verbose)
    return {name: _get_expr(dt, val) for name, (dt, val) in variables.items()}


def evaluate_c_defined_expressions(variables: dict[str, Expr], initialized_in: str, verbose=False) -> dict[str, Expr]:
    """For defined identifiers, evaluate a (simple) block of C code to determine the end values of the identifiers"""
    names_types = {name: expr.data_type.name for name, expr in variables.items()}
    return evaluate_c_defined_variables(names_types, initialized_in, known=variables, verbose=verbose)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Union
from pathlib import Path
from numpy import ndarray


@dataclass
class DatFileCommon:
    source: Path
    metadata: dict = field(default_factory=dict)
    parameters: dict = field(default_factory=dict)
    variables: list[str] = field(default_factory=list)
    data: ndarray = field(default_factory=ndarray)

    @classmethod
    def from_filename(cls, filename: str | Path):
        from numpy import array
        source = Path(filename).resolve()
        if not source.exists():
            raise RuntimeError('Source filename does not exist')
        if not source.is_file():
            raise RuntimeError(f'{filename} does not name a valid file')
        with source.open('r') as file:
            lines = file.readlines()

        header = [x.strip(' #\n') for x in filter(lambda x: x[0] == '#', lines)]
        meta = {k.strip(): v.strip() for k, v in
                [x.split(':', 1) for x in filter(lambda x: not x.startswith('Param'), header)]}
        parm = {k.strip(): v.strip() for k, v in
                [x.split(':', 1)[1].split('=', 1) for x in filter(lambda x: x.startswith('Param'), header)]}
        var = meta.get('variables', '').split(' ')
        data = array([[float(x) for x in line.strip().split()] for line in filter(lambda x: x[0] != '#', lines)])
        return cls(source, meta, parm, var, data)

    @staticmethod
    def parts():
        return [], []

    def print_data(self, file):
        pass

    def to_filename(self, filename: str):
        sink = Path(filename).resolve()
        if sink.exists():
            raise RuntimeError(f'{filename} already exists')
        first, second = self.parts()
        with sink.open('w') as file:
            for item in first:
                print(f'# {item}: {self.metadata[item]}', file=file)
            for param in self.parameters:
                print(f'# Param: {param}={self.parameters[param]}', file=file)
            for item in second:
                print(f'# {item}: {self.metadata[item]}', file=file)
            self.print_data(file)

    def __getitem__(self, item):
        if item in self.variables:
            index = [i for i, x in enumerate(self.variables) if x == item]
            if len(index) != 1:
                raise RuntimeError(f'Expected one index for {item} but found {index}')
            return self.data[index[0], ...]
        elif item in self.parameters:
            return self.parameters[item]
        elif item in self.metadata:
            return self.metadata[item]
        else:
            raise KeyError(f'Unknown key {item}')

    def dim_metadata(self) -> list[dict]:
        pass

    @staticmethod
    def safe_to_combine(other):
        return False

    def __add__(self, other):
        if not self.safe_to_combine(other):
            raise RuntimeError('Cannot combine these two dat files')
        metadata = combine_scan_dicts(self.metadata, other.metadata)
        parameters = combine_scan_dicts(self.parameters, other.parameters)
        variables = combine_scan_lists(self.variables, other.variables)
        data = combine_scan_data(self.data, other.data)
        return DatFileCommon(Path(), metadata, parameters, variables, data)


def dim_metadata(length, label_unit, lower_limit, upper_limit) -> dict:
    from numpy import linspace
    label = label_unit.split(' ', 1)[0]
    unit = label_unit.split(' ', 1)[1].strip('[] ')
    if '\\gms' == unit:
        unit = 'microseconds'
    bin_width = (upper_limit - lower_limit) / (length - 1)
    boundaries = linspace(lower_limit - bin_width / 2, upper_limit + bin_width / 2, length + 1)
    return dict(lenght=length, label=label, unit=unit, bin_boundaries=boundaries)


@dataclass
class DatFile0D(DatFileCommon):
    def __post_init__(self):
        nv = len(self.variables)
        if self.data.shape == (nv, ):
            # shortcut in case we're already in the right shape (e.g., from __add__)
            return
        if self.data.size != nv:
            raise RuntimeError(f'Unexpected data shape {self.data.shape} for metadata specifying {nv=}')
        self.data = self.data.reshape((nv, ))

    def dim_metadata(self) -> list[dict]:
        return []

    def print_data(self, file):
        print(' '.join(str(x) for x in self.data), file=file)

    @staticmethod
    def parts():
        # Yes, Ncount shows up twice ...
        first = ('Format', 'URL', 'Creator', 'Instrument', 'Ncount', 'Trace', 'Gravitation', 'Seed', 'Directory')
        second = ('Date', 'type', 'Source', 'component', 'position', 'title', 'Ncount', 'filename', 'statistics',
                  'signal', 'values', 'yvar', 'ylabel', 'xlimits', 'variables')
        return first, second

    def safe_to_combine(self, other):
        if not isinstance(other, DatFile1D):
            return False
        if self.variables != other.variables:
            return False
        if self.data.shape != other.data.shape:
            return False
        return True

    def __add__(self, other):
        both = super().__add__(other)
        return DatFile1D(both.source, both.metadata, both.parameters, both.variables, both.data)


@dataclass
class DatFile1D(DatFileCommon):
    def __post_init__(self):
        nx = int(self.metadata['type'].split('(', 1)[1].strip(')'))
        nv = len(self.variables)
        if self.data.shape == (nv, nx):
            # shortcut in case we're already in the right shape (e.g., from __add__)
            return
        if self.data.shape[0] != nx or self.data.shape[1] != nv:
            raise RuntimeError(f'Unexpected data shape {self.data.shape} for metadata specifying {nx=} and {nv=}')
        # we always want the variables along the first dimension:
        self.data = self.data.transpose((1, 0))

    def dim_metadata(self) -> list[dict]:
        lower_limit, upper_limit = [float(x) for x in self['xlimits'].split()]
        return [dim_metadata(self.data.shape[1], self['xlabel'], lower_limit, upper_limit), ]

    def print_data(self, file):
        for row in self.data.transpose((1, 0)):
            print(' '.join(str(x) for x in row), file=file)

    @staticmethod
    def parts():
        # Yes, Ncount shows up twice ...
        first = ('Format', 'URL', 'Creator', 'Instrument', 'Ncount', 'Trace', 'Gravitation', 'Seed', 'Directory')
        second = ('Date', 'type', 'Source', 'component', 'position', 'title', 'Ncount', 'filename', 'statistics',
                  'signal', 'values', 'xvar', 'yvar', 'xlabel', 'ylabel', 'xlimits', 'variables')
        return first, second

    def safe_to_combine(self, other):
        if not isinstance(other, DatFile1D):
            return False
        if self.variables != other.variables:
            return False
        if self.metadata['xlabel'] != other.metadata['xlabel']:
            return False
        if self.metadata['xlimits'] != other.metadata['xlimits']:
            return False
        if self.data.shape != other.data.shape:
            return False
        return True

    def __add__(self, other):
        both = super().__add__(other)
        return DatFile1D(both.source, both.metadata, both.parameters, both.variables, both.data)


@dataclass
class DatFile2D(DatFileCommon):
    def __post_init__(self):
        nx, ny = [int(x) for x in self.metadata['type'].split('(', 1)[1].strip(')').split(',')]
        nv = len(self.variables)
        if self.data.shape == (nv, ny, nx):
            # shortcut in case we're already in the right shape (e.g., from __add__)
            return
        if self.data.shape[0] != ny * nv or self.data.shape[1] != nx:
            raise RuntimeError(f'Expected {ny*nv =} by {nx =} but have {self.data.shape}')
        self.data = self.data.reshape((nv, ny, nx))

    def dim_metadata(self) -> list[dict]:
        lower_x, upper_x, lower_y, upper_y = [float(x) for x in self['xylimits'].split()]
        return [dim_metadata(self.data.shape[2], self['xlabel'], lower_x, upper_x),
                dim_metadata(self.data.shape[1], self['ylabel'], lower_y, upper_y)]

    @staticmethod
    def parts():
        first = ('Format', 'URL', 'Creator', 'Instrument', 'Ncount', 'Trace', 'Gravitation', 'Seed', 'Directory')
        second = ('Date', 'type', 'Source', 'component', 'position', 'title', 'Ncount', 'filename', 'statistics',
                  'signal', 'values', 'xvar', 'yvar', 'xlabel', 'ylabel', 'zvar', 'zlabel', 'xylimits', 'variables')
        return first, second

    def print_data(self, file):
        labels = 'Data', 'Errors', 'Events'  # McPlot looks at only these labels to determine what is the data
        for i, n in enumerate(self.variables):
            print(f'# {labels[i]} [{self.metadata["component"]}/{self.metadata["filename"]}] {n}:', file=file)
            for row in self.data[i, ...]:
                print(' '.join(str(x) for x in row), file=file)

    def safe_to_combine(self, other):
        if not isinstance(other, DatFile2D):
            return False
        if self.variables != other.variables:
            return False
        if self.metadata['xlabel'] != other.metadata['xlabel']:
            return False
        if self.metadata['ylabel'] != other.metadata['ylabel']:
            return False
        if self.metadata['xylimits'] != other.metadata['xylimits']:
            return False
        if self.data.shape != other.data.shape:
            return False
        return True

    def __add__(self, other):
        both = super().__add__(other)
        return DatFile2D(both.source, both.metadata, both.parameters, both.variables, both.data)


def read_mccode_dat(filename: str | Path):
    common = DatFileCommon.from_filename(filename)
    array_type = common.metadata['type']
    ndim, data_type = -1, None
    if array_type.startswith('array_0d'):
        ndim, data_type = 0, DatFile0D
    elif array_type.startswith('array_1d'):
        ndim, data_type = 1, DatFile1D
    elif array_type.startswith('array_2d'):
        ndim, data_type = 2, DatFile2D
    if ndim < 0 or ndim > 2:
        raise RuntimeError(f'Unexpected number of dimensions: {ndim}')
    return data_type(common.source, common.metadata, common.parameters, common.variables, common.data)


def combine_scan_dicts(a: dict, b: dict):
    from copy import deepcopy
    c = deepcopy(a)
    special_add = 'Ncount',
    special_concatenate = 'Directory', 'filename', 'Date', 'Seed'
    special_ignore = 'signal', 'statistics', 'values'
    for k, v in b.items():
        if any(s in k for s in special_add):
            c[k] = int(c[k]) + int(v)
        elif any(s in k for s in special_concatenate):
            if v not in c[k]:
                c[k] = f'{c[k]} {v}'
        elif any(s in k for s in special_ignore):
            pass
        elif k in c:
            if c[k] != v:
                raise RuntimeError(f'Incompatible values for "{k}": {c[k]} and {v}')
        else:
            raise RuntimeError(f'Unexpected key {k} in {b}')
    return c


def combine_scan_lists(a: list, b: list):
    if a != b:
        raise RuntimeError(f'Incompatible lists {a} and {b}')
    return a


def combine_scan_data(a: ndarray, b: ndarray):
    if a.shape != b.shape:
        raise RuntimeError(f'Incompatible shapes {a.shape} and {b.shape}')
    if a.ndim == 2:
        from copy import deepcopy
        # variables are _always_ (?) (..., I, I_err, N)
        c = deepcopy(a)
        c[-3:, :] = a[-3:, :] + b[-3:, :]
        return c
    elif a.ndim == 3:
        # but 3-D _only_ contains (I, I_err, N)
        return a + b
    else:
        raise RuntimeError(f'Unexpected number of dimensions: {a.ndim}')


def combine_mccode_dats(dats: list):
    one = dats[0]
    for other in dats[1:]:
        one = one + other
    return one


def write_combined_mccode_dats(files: list[Union[Path, str]], output: Union[Path, str]):
    dats = combine_mccode_dats([read_mccode_dat(f) for f in files])
    dats.to_filename(output)





from typing import Union
from pathlib import Path
from zenlog import log
from dataclasses import dataclass, field
from antlr4.error.ErrorListener import ErrorListener
from .registry import Registry, MCSTAS_REGISTRY, registries_match, registry_from_specification
from ..comp import Comp


class ReaderErrorListener(ErrorListener):
    def __init__(self, filetype: str, name: str, source: str, pre=5, post=2):
        self.filetype = filetype
        self.name = name
        self.source = source
        self.pre = pre
        self.post = post

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        log.error(f'Syntax error in parsing {self.filetype} {self.name} at {line},{column}')
        lines = self.source.split('\n')
        pre_lines = lines[line-self.pre:line]
        post_lines = lines[line:line+self.post]
        for line in pre_lines:
            log.info(line)
        log.error('~'*column + '^ ' + msg)
        for line in post_lines:
            log.info(line)


@dataclass
class Reader:
    registries: list[Registry] = field(default_factory=list)
    components: dict[str, Comp] = field(default_factory=dict)
    c_flags: list[str] = field(default_factory=list)

    def __post_init__(self):
        if len(self.registries) == 0:
            self.registries = [MCSTAS_REGISTRY, ]

    def prepend_registry(self, reg: Registry):
        self.registries[:0] = [reg, ]

    def append_registry(self, reg: Registry):
        self.registries.append(reg)

    def handle_search_keyword(self, spec: str):
        if not any(registries_match(reg, spec) for reg in self.registries):
            reg = registry_from_specification(spec)
            if reg is not None:
                self.prepend_registry(reg)
            else:
                raise RuntimeError(f"Registry specification {spec} did not specify a valid registry!")

    def add_c_flags(self, flags):
        self.c_flags.append(flags)

    def locate(self, name: str, which: str = None, ext: str = None):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        for reg in registries:
            if reg.known(name, ext):
                return reg.path(name, ext)
        names = [reg.name for reg in registries]
        msg = "registry " + names[0] if len(names) == 1 else 'registries: ' + ','.join(names)
        raise RuntimeError(f'{name} not found in {msg}')

    def fullname(self, name: str, which: str = None, ext: str=None):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        for reg in registries:
            if reg.known(name, ext):
                return reg.fullname(name, ext)
        names = [reg.name for reg in registries]
        msg = "registry " + names[0] if len(names) == 1 else 'registries: ' + ','.join(names)
        raise RuntimeError(f'{name} not found in {msg}')

    def known(self, name: str, which: str = None):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        return any([reg.known(name) for reg in registries])

    def unique(self, name: str, which: str = None):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        return sum([1 for reg in registries if reg.unique(name)]) == 1

    def contain(self, name: str, which: str = None):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        return [reg.name for reg in registries if reg.known(name)]

    def stream(self, name: str, which: str = None):
        from antlr4 import FileStream
        return FileStream(str(self.locate(name, which=which).resolve()), encoding='utf8')

    def add_component(self, name: str, current_instance_name=None):
        if name in self.components:
            raise RuntimeError("The named component is already known.")
        from antlr4 import CommonTokenStream, FileStream
        from ..grammar import McCompLexer, McCompParser
        from ..comp import CompVisitor
        filename = str(self.locate(name, ext='.comp').resolve())
        with open(filename, 'r') as file:
            source = file.read()
        lexer = McCompLexer(FileStream(filename, encoding='utf8'))
        tokens = CommonTokenStream(lexer)
        parser = McCompParser(tokens)
        parser.addErrorListener(ReaderErrorListener('Component', name, source))
        visitor = CompVisitor(self, filename, instance_name=current_instance_name)  # The visitor needs to be able to call *this* method
        res = visitor.visitProg(parser.prog())
        if not isinstance(res, Comp):
            raise RuntimeError(f'Parsing component file {filename} did not produce a component object!')
        if res.category is None:
            # guess the component category from the registered filename (not fully resolved path)
            fullname = self.fullname(name, ext='.comp')
            fullname = fullname if isinstance(fullname, Path) else Path(fullname)
            # if fullname is an absolute path, it comes from a local repository -- so we don't know what to do
            res.category = 'UNKNOWN' if fullname.is_absolute() else fullname.parts[0]
        self.components[name] = res

    def get_component(self, name: str, current_instance_name=None):
        if name not in self.components:
            self.add_component(name, current_instance_name=current_instance_name)
        return self.components[name]

    def get_instrument(self, name: Union[str, Path], destination=None):
        """Load and parse an instr Instrument definition file

        In McCode3 fashion, the instrument file *should* be in the current working directory.
        In new-fashion, the registry/registries will be checked if it is not.
        """
        from antlr4 import CommonTokenStream, FileStream
        from ..grammar import McInstrParser, McInstrLexer
        from ..instr import InstrVisitor, Instr
        path = name if isinstance(name, Path) else Path(name)
        if path.suffix != '.instr':
            path = path.with_suffix(f'{path.suffix}.instr')
        if not path.exists() and not path.is_file():
            path = self.locate(path.name)  # include the .instr for the search
        if not path.exists() and not path.is_file():
            raise RuntimeError(f'Can not locate instr file for {name}.')
        filename = str(path.resolve())
        with open(filename, 'r') as file:
            source = file.read()
        lexer = McInstrLexer(FileStream(filename, encoding='utf8'))
        tokens = CommonTokenStream(lexer)
        parser = McInstrParser(tokens)
        parser.addErrorListener(ReaderErrorListener('Instrument', name, source))
        visitor = InstrVisitor(self, filename, destination=destination)
        res = visitor.visitProg(parser.prog())
        if not isinstance(res, Instr):
            raise RuntimeError(f'Parsing instrument file {filename} did not produce an Instr object')
        res.source = filename
        res.flags = tuple(self.c_flags)
        res.registries = tuple(self.registries)
        return res
